import random
import re
import networkx as nx

####################     Reporter     ####################
class Reporter(object):
    def __init__(self):
        self.error = False
        self.mess = ""
        pass
        
    def display(self,  mes):
        self.error = True
        mess = self.mess + "ERROR -> "+mes
        trace_file.write("\n\n"+ mess)
        
####################     MacTreeRandom Object for test     ####################        

class MacTreeRandom(object):
    def __init__(self, name, frequency, lcam, random_list, deep, tag_name_list):
        self.name = name
        self.frequency = frequency
        
        self.deep = -1
        self.tag_name = ''
        
        self.related_graph = None
        
        if name :
            self.deep = deep
            count=0
            tag_name = '__'+str(count)+'__'+self.name+'_'+str(self.frequency)
            while tag_name in tag_name_list :
                count+=1
                tag_name = '__'+str(count)+'__'+self.name+'_'+str(self.frequency)
            self.tag_name = tag_name
            tag_name_list.append(self.tag_name)
            
        self.children = []
        
        while len(lcam)>0:                   
            # random choice  
            index_cam = random.randint(0,len(lcam)-1)
            for elemt in random_list :
                if elemt in lcam[index_cam]:
                    max_mol = elemt
                    break                    
            max_freq = 10
            

            # cam lists for next child
            lcam_mol = []
            new_lcam = []
            for cam in lcam:
                if max_mol in cam:
                    cam.remove(max_mol)
                    if len(cam)>0:
                        lcam_mol.append(cam)
                else:
                    new_lcam.append(cam)
            
            new_child = MacTreeRandom(max_mol, max_freq, lcam_mol, random_list, deep+1, tag_name_list)
            self.children.append(new_child)
            lcam = new_lcam
    
    
    def to_di_graph(self, dg, father, freq_dict):
        if self.name:
            name = self.tag_name
            try :
                freq = freq_dict[flatName2name(name)]
            except :
                print 'Unknown name : ',flatName2name(name)
                freq = 10
            dg.add_node(name, size = freq, label = self.name)
            if father:
                dg.add_edges_from([(father,name)])
            father = name
            
        for ct in self.children:
            ct.to_di_graph(dg, father, freq_dict)
        return dg    
    

    def set_related_graph(self,graph):
        self.related_graph = graph

        
    def make_deep_dict(self,deep_dict):
        """
        return tree_dict : key = depth (int), value = list of CamTree Object
        """
        if self.name :
            if deep_dict.has_key(self.deep):
                deep_dict[self.deep].append(self)            
            else :
                deep_dict[self.deep]=[self]
        for child in self.children:
            child.make_deep_dict(deep_dict) 
        return deep_dict
    
    def to_tuple(self):
        cam_tuple = []
        for child in self.children :
            mini_graph = child.related_graph
            cam_tuple.append(mini_graph)   
        cam_tuple.sort()
        cam_tuple.insert(0,self.name)
        return tuple(cam_tuple)



####################     MacTree Object     ####################

class MacTree(object):
    def __init__(self, name, frequency, lcam, distance_dict, deep, tag_name_list, freqFirst=True, far=True):
        self.name = name
        self.frequency = frequency
        self.distance = -1
        self.deep = -1
        self.tag_name = ''
        
        self.related_graph = None
        
        if name :
            self.distance = distance_dict[name]
            self.deep = deep
            count=0
            tag_name = '__'+str(count)+'__'+self.name+'_'+str(self.frequency)
            while tag_name in tag_name_list :
                count+=1
                tag_name = '__'+str(count)+'__'+self.name+'_'+str(self.frequency)
            self.tag_name = tag_name
            tag_name_list.append(self.tag_name)
            
        self.children = []
        
        while len(lcam)>0:
            if freqFirst:
                # frequency and distance computation
                mol_list,max_freq = get_most_freq_mol(lcam)
                if far :
                    mol_list = get_farest_mol(mol_list,distance_dict)
                else :
                    mol_list = get_nearest_mol(mol_list,distance_dict)
                mol_list.sort()
                max_mol = mol_list[0]


            else :
               best_dist_lcam=[]
               for cam in lcam:
                    if far :
                        best_dist_in_cam = get_farest_mol(cam,distance_dict)
                    else :
                        best_dist_in_cam = get_nearest_mol(cam,distance_dict)
                    best_dist_lcam.append(best_dist_in_cam)
                    mol_list,max_freq = get_most_freq_mol(best_dist_lcam)
                    mol_list.sort()
                    max_mol = mol_list[0]
                    
            # random choice
            """
            index_cam = random.randint(0,len(lcam)-1)
            index_mol = random.randint(0,len(lcam[index_cam])-1)
            max_mol = lcam[index_cam][index_mol]
            max_freq = 10
            """

            # cam lists for next child
            lcam_mol = []
            new_lcam = []
            for cam in lcam:
                if max_mol in cam:
                    cam.remove(max_mol)
                    if len(cam)>0:
                        lcam_mol.append(cam)
                else:
                    new_lcam.append(cam)
#            print lcam
#            print max_mol
#            print
            new_child = MacTree(max_mol, max_freq, lcam_mol, distance_dict, deep+1, tag_name_list, freqFirst, far)
            self.children.append(new_child)
            lcam = new_lcam
    
    
    def to_di_graph(self, dg, father, freq_dict):
        if self.name:
            name = self.tag_name
            try :
                freq = freq_dict[flatName2name(name)]
            except :
                print 'Unknown name : ',flatName2name(name)
                freq = 10
            dg.add_node(name, size = freq, label = self.name)
            if father:
                dg.add_edges_from([(father,name)])
            father = name
            
        for ct in self.children:
            ct.to_di_graph(dg, father, freq_dict)
        return dg    
    

    def set_related_graph(self,graph):
        self.related_graph = graph

        
    def make_deep_dict(self,deep_dict):
        """
        return tree_dict : key = depth (int), value = list of CamTree Object
        """
        if self.name :
            if deep_dict.has_key(self.deep):
                deep_dict[self.deep].append(self)            
            else :
                deep_dict[self.deep]=[self]
        for child in self.children:
            child.make_deep_dict(deep_dict) 
        return deep_dict
    
    def to_tuple(self):
        cam_tuple = []
        for child in self.children :
            mini_graph = child.related_graph
            cam_tuple.append(mini_graph)   
        cam_tuple.sort()
        cam_tuple.insert(0,self.name)
        return tuple(cam_tuple)
    
####################     MAC DAG Object     ####################
    
class MacDag(object):
    def __init__(self, cam_tree_obj, child_list):
        self.name = cam_tree_obj.name
        self.tag_name = cam_tree_obj.tag_name
        self.equivalent_node_list = []
        self.parents= []
        self.children = child_list
        
        self.linear_counter = 1
        self.horizontal_counter = 1
        
    def pretty_print(self):
        if self.name:
            print self
            print 'name : ',self.name
            print 'tag name : ',self.tag_name
            print 'parents : ', self.parents
            for p in self.parents :
                print p.tag_name
            print 'children : ',self.children
            for c in self.children :
                print c.tag_name
            print
        for child in self.children :
            child.pretty_print()
        
    def add_equivalent_node(self,node_name):
        if node_name not in self.equivalent_node_list :
            self.equivalent_node_list.append(node_name)
    
    def replace_parent(self, old_parent, new_parent):
        for parent in self.parents :
            if parent == old_parent :
                self.parents.remove(parent)
        self.parents.append(new_parent)
        
    def remove_parent(self, parent_to_remove):
        for parent in self.parents :
            if parent == parent_to_remove :
                self.parents.remove(parent)
        
    def remove_child(self, child_to_remove):
        for child in self.children :
            if child == child_to_remove :
                self.children.remove(child)
  
    def allocate_father(self,father = None):
        if self.name:
            if father and father not in self.parents :
                self.parents.append(father)
        for child in self.children : 
            child.allocate_father(self)
            
            
    def linear_reduction(self, already_visited_list = [], reduc_dict = None):
        already_visited = already_visited_list
        reduction_dict = reduc_dict
        # Make reduction_dict the first time
        if not reduction_dict :
            reduction_dict = dict()
            reduction_dict = self.make_reduction_dict(reduction_dict, None)
            for key in reduction_dict.keys(): # remove non linear
                if len(reduction_dict[key]) == 1:
                    del reduction_dict[key]

        if self.name and self.tag_name not in already_visited :
            if reduction_dict.has_key(self.tag_name):
                already_visited.append(self.tag_name)
                old_parent = reduction_dict[self.tag_name][-1]
                (self.linear_counter, self.horizontal_counter) = self.get_merged_linear_counter(reduction_dict[self.tag_name])
                self.name,self.tag_name,self.children = self.get_new_children(reduction_dict, self.tag_name)

#                print self.name # display new name
                for child in self.children :
                    child.replace_parent(old_parent, self)
        for child in self.children :
            child.linear_reduction(already_visited, reduction_dict)
            
            
    def get_new_children(self, reduction_dict, origin_name):
        origin_dag = reduction_dict[origin_name][0]
        merge_name = '(_MERGE_'+origin_dag.name
        merge_tag_name = '(_MERGE_'+origin_dag.tag_name
        
        for linear_dag in reduction_dict[origin_name][1:]:
            merge_name += '_AND_'+linear_dag.name
            merge_tag_name += '_AND_'+linear_dag.tag_name
        merge_name += ')'
        merge_tag_name += ')'
        
        origin_dag.name = merge_name
        origin_dag.tag_name = merge_tag_name
        origin_dag.children = reduction_dict[origin_name][-1].children
        return (origin_dag.name, origin_dag.tag_name, origin_dag.children)
            
    def make_reduction_dict(self, reduction_dict, key):
        if self.name :
            if not key :
#                print 'KEY name : ',self.name, 'father : ',self.parents
                if len(self.children) == 1:
                    key = self.tag_name
                    reduction_dict = self.manage_reduction_dict(reduction_dict, key, self)
            else :  
                if len(self.children) <= 1 and len(self.parents) == 1 :
                    reduction_dict = self.manage_reduction_dict(reduction_dict, key, self)
                elif len(self.parents) == 1 :
                    reduction_dict = self.manage_reduction_dict(reduction_dict, key, self)
                    key = None
                    self.make_reduction_dict(reduction_dict, key)
                else :
                    key = None
                    self.make_reduction_dict(reduction_dict, key)
        
        for child in self.children :
            child.make_reduction_dict(reduction_dict, key)
        return  reduction_dict
    
    def manage_reduction_dict(self,dictionary,key,value):
        if dictionary.has_key(key) :
            if value not in dictionary[key] :
                dictionary[key].append(value)
        else :
            dictionary[key]=[value]
        return dictionary
    
    def horizontal_reduction(self):
        local_env_dict = dict() # MUST start with empty dict
        local_env_dict = self.make_local_env_dict(local_env_dict)
        for key in local_env_dict.keys():
                if len(local_env_dict[key]) == 1:
                    del local_env_dict[key]
        
        while len(local_env_dict.keys()) != 0 :
            first_key = local_env_dict.keys()[0]
            reference_dag = local_env_dict[first_key][0]

            # MAKE NEW NAME
            new_name = '('+reference_dag.name
            new_tag_name = '('+reference_dag.tag_name
            for other_dag in local_env_dict[first_key][1:]:
                new_name += '_OR_'+ other_dag.name
                new_tag_name += '_OR_'+ other_dag.tag_name
            new_name += ')'
            new_tag_name += ')'
#            print new_name #display new name
            
            # CHANGE ATTRIBUTES
            reference_dag.name = new_name
            reference_dag.tag_name = new_tag_name
            reference_dag.horizontal_counter, reference_dag.linear_counter = self.get_merged_horizontal_counter(local_env_dict[first_key])
            for parent in reference_dag.parents :
                for dag in local_env_dict[first_key][1:] :
                    parent.remove_child(dag)
                
            for child in reference_dag.children :
                for dag in local_env_dict[first_key][1:] :
                    child.remove_parent(dag)            
            
            local_env_dict = dict() # MUST start with empty dict
            local_env_dict = self.make_local_env_dict(local_env_dict)
            for key in local_env_dict.keys():
                if len(local_env_dict[key]) <= 1:
                    del local_env_dict[key]

    
    def make_local_env_dict(self, local_env_dict):
        parents_list = []
        children_list = []
        for par in self.parents :
            parents_list.append(par.tag_name)
        for child in self.children : 
            children_list.append(child.tag_name)
        parents_list.sort()
        children_list.sort()
        if len(parents_list) == 0 :
            local_env = [None]+children_list
        elif len(children_list) == 0:
            local_env = parents_list+[None]
        else :
            local_env = parents_list + children_list
        local_env = tuple(local_env)
        if local_env_dict.has_key(local_env) :
            if self not in local_env_dict[local_env] :
                local_env_dict[local_env].append(self)
        else :
            local_env_dict[local_env] = [self]
            
        for child in self.children :
            child.make_local_env_dict(local_env_dict)
        return local_env_dict
        
            
    def to_di_graph(self, dg, father, freq_dict):
        if self.name:
            name = self.tag_name
            try :
                freq = freq_dict[flatName2name(name)]
            except :
                freq = 10
            dg.add_node(name,size=freq, label = self.name)
            if father:
                dg.add_edges_from([(father,name)])
            father = name
            
        for ct in self.children:
            ct.to_di_graph(dg, father, freq_dict)
        return dg
    
    
    def to_digraph_merge_representation(self, dg, father):
        if self.name:
            name = self.tag_name          
            freq = self.linear_counter*10

            dg.add_node(name, size=freq, label = self.name)
            if father:
                dg.add_edges_from([(father,name)])
            father = name
            
        for ct in self.children:
            ct.to_digraph_merge_representation(dg, father)
        return dg
    
    def get_merged_linear_counter(self,list_of_dag):
        merged_lc = 0
        merged_hc = 0 #take care previous horizontal merge
        for dag in list_of_dag :
            merged_lc += dag.linear_counter  
            if  dag.horizontal_counter > 1 :
                merged_hc += dag.horizontal_counter
        if merged_hc == 0 :
            merged_hc = 1
        return (merged_lc, merged_hc)
    
    def get_merged_horizontal_counter(self,list_of_dag):
        merged_hc = 0
        merged_lc = 0 #take care previous linear merge
        for dag in list_of_dag :
            merged_hc += dag.horizontal_counter  
            if dag.linear_counter > 1 :
                merged_lc += dag.linear_counter
        if merged_lc == 0:
            merged_lc = 1      
        return (merged_hc,merged_lc)
    
####################     MAC DAG Object Builder    ####################    

def build(deep_dict):
    list_of_cam_dag = []
    cam_tuple_dict = dict()
    depth = get_max_key(deep_dict)
    while depth > -1 :
        list_of_cam_tree = deep_dict[depth] 
        for cam_tree in list_of_cam_tree :
            cam_tuple = cam_tree.to_tuple()
            if cam_tuple_dict.has_key(cam_tuple) :
                cam_dag = cam_tuple_dict[cam_tuple]
            else :
                if len(cam_tuple) == 1 :
                    cam_dag = MacDag(cam_tree, [])      
                else :
                    child_list = list(cam_tuple[1:])
                    cam_dag = MacDag(cam_tree, child_list)
                cam_tuple_dict[cam_tuple] = cam_dag
              
            cam_dag.add_equivalent_node(cam_tree.tag_name)
            cam_tree.set_related_graph(cam_dag)
            if depth == 0 :
                list_of_cam_dag.append(cam_dag)
        depth -= 1
    return list_of_cam_dag
    
####################     Functions     ####################

# Random test
def make_random_list(input_list):
    order_list = input_list
    random_list = []
    while len(order_list) != 0:
        index = random.randint(0,len(order_list)-1)
        elemt = order_list[index]
        random_list.append(elemt)
        order_list.remove(elemt)      
    return random_list

# Dictionary of property predecessors
def makePredDict(chart_file, lprop):
        """
        

        """
        report = Reporter()

        sa = StaticAnalyzer(report)
        sa.build_from_chart_file(chart_file)
#        graph=sa.make_full_dependence_dg()
        graph=sa.make_dependence_dg()
        pred_dict = {}
        
        # SEARCH MAXIMAL DISTANCE
        for prop in lprop :
            pred_list = sa.get_predecessors(graph, prop)

            for pred in pred_list:
                    dist = nx.shortest_path_length(graph,source=pred,target=prop)
                    if  pred_dict.has_key(pred) :
                        if pred_dict[pred]>dist: # we take the farthest 
                            pred_dict[pred]=dist
                    else :
                        pred_dict[pred]=dist
        return pred_dict
#        report = Reporter()
#
#        sa = StaticAnalyzer(report)
#        sa.build_from_chart_file(chart_file)
#        graph=sa.make_dependence_dg()
#        graph = graph.to_undirected()
#        
#        pred_dict = {}
#        
#        # SEARCH MAXIMAL DISTANCE
#        for prop in lprop :
#            pred_list = graph.nodes()
#            for pred in pred_list:
#                    dist = nx.shortest_path_length(graph,source=pred,target=prop)
#                    if  pred_dict.has_key(pred) :
#                        if pred_dict[pred]>dist: # we take the farthest 
#                            pred_dict[pred]=dist
#                    else :
#                        pred_dict[pred]=dist
#        return pred_dict
            
    
# Order MAC elements
def get_most_freq_mol(list_of_mac):
    freq = dict()
    for cam in list_of_mac:
        for mol in cam:
            try:
                fr = freq[mol]
                freq[mol] = fr + 1
            except KeyError:
                freq[mol] = 1
    # most frequent molecule
    max_freq = 0
    for mol in freq.keys():
        f = freq[mol]
        if f > max_freq:
            max_freq = f
            
    max_mol_list=[]
    for mol in freq.keys():
        f = freq[mol]
        if f == max_freq :
            max_mol_list.append(mol)
    return (max_mol_list,max_freq)

def get_farest_mol(list_of_name, distance_dict):
    far_list=[]
    for key in list_of_name:
        if len(far_list)==0:
            far_list.append(key)
        elif distance_dict[key]>distance_dict[far_list[0]]:
            far_list=[key]
        elif distance_dict[key]==distance_dict[far_list[0]]:
            far_list.append(key)
    return far_list

def get_nearest_mol(list_of_name, distance_dict):
    near_list=[]
    for key in list_of_name:
        if len(near_list)==0:
            near_list.append(key)
        elif distance_dict[key]<distance_dict[near_list[0]]:
            near_list=[key]
        elif distance_dict[key]==distance_dict[near_list[0]]:
            near_list.append(key)
    return near_list

def flatName2name(name):
    name=re.sub(r"__[0-9]+__", r"",name)
    name=re.sub(r"_[0-9]+$", r"",name)
    return name


# Dictionary Functions

def make_freq_dict(list_of_mac):
    freq_dict = dict()
    for mac in list_of_mac :
        for elemt in mac :
            if freq_dict.has_key(elemt):
                freq_dict[elemt]+=1
            else :
                freq_dict[elemt]=1
    return freq_dict

def get_max_value(dictionary):
    max_value=0
    for key in dictionary.keys():
        if dictionary[key]>max_value:
            max_value=dictionary[key]
    return max_value

def get_max_key(dictionary):
    max_key=0
    for key in dictionary.keys():
        if key>max_key:
            max_key=key
    return max_key

def normalise_value(dictionary, norm_val):
    max_val = get_max_value(dictionary)
    new_dict=dict()
    for key in dictionary.keys():
        new_value = (dictionary[key] * norm_val)/max_val*1.0
        new_dict[key]=new_value
    return new_dict   


# MAC functions
def mac_file_2_list(mac_input_file):
    cam_list = []
    lm = mac_input_file.readline()
    while lm:
        cam_list.append(lm[:-1].split())
        lm = mac_input_file.readline()
    return cam_list

def mac_distance_filter(mac, dist_dict, min, max):
    new_mac = []
    for elemt in mac :
        if dist_dict[elemt] >= min and dist_dict[elemt]<=max :
            new_mac.append(elemt)  
    return new_mac

def mac_list_2_element_list(mac_list): # for random test
    element_list = []
    for mac in mac_list :
        for elemt in mac :
            if elemt not in element_list :
                element_list.append(elemt)
    return element_list
    
# Display functions
def red2blue_10grad(graph, normalized_dict):
    for n in graph.nodes():
        node = graph.node[n]
        name = node['label']
        dist = normalized_dict[name]
        
#        name = flatName2name(n.label)
#        dist = normalized_dict[name]
        if dist <=10 :
            graph.node[n]['r'] = 194
	    graph.node[n]['g'] = 11
	    graph.node[n]['b'] = 84
        elif dist <=20 :
            graph.node[n]['r'] = 211
	    graph.node[n]['g'] = 36
	    graph.node[n]['b'] = 108
        elif dist <=30 :
            graph.node[n]['r'] = 228
	    graph.node[n]['g'] = 62
	    graph.node[n]['b'] = 132
        elif dist <=40 :
            graph.node[n]['r'] = 245
	    graph.node[n]['g'] = 88
	    graph.node[n]['b'] = 156
        elif dist <=50 :
            graph.node[n]['r'] = 201
	    graph.node[n]['g'] = 105
	    graph.node[n]['b'] = 187
        elif dist <=60 :
            graph.node[n]['r'] = 158
	    graph.node[n]['g'] = 122
	    graph.node[n]['b'] = 218
        elif dist <=70 :
            graph.node[n]['r'] = 115
	    graph.node[n]['g'] = 140
	    graph.node[n]['b'] = 250
        elif dist <=80 :
            graph.node[n]['r'] = 76
	    graph.node[n]['g'] = 110
	    graph.node[n]['b'] = 241
        elif dist <=90 :
            graph.node[n]['r'] = 38
	    graph.node[n]['g'] = 81
	    graph.node[n]['b'] = 232
        else :
            graph.node[n]['r'] = 0
	    graph.node[n]['g'] = 52
	    graph.node[n]['b'] = 224

def red2blue_5grad(graph, normalized_dict):
    for n in graph.nodes():
        node = graph.node[n]
        name = node['label']
        dist = normalized_dict[name]
        if dist <=10 :
            graph.node[n]['r'] = 255
	    graph.node[n]['g'] = 204
	    graph.node[n]['b'] = 230
        elif dist <=20 :
            graph.node[n]['r'] = 242
	    graph.node[n]['g'] = 216
	    graph.node[n]['b'] = 236
        elif dist <=30 :
            graph.node[n]['r'] = 229
	    graph.node[n]['g'] = 229
	    graph.node[n]['b'] = 242
        elif dist <=40 :
            graph.node[n]['r'] = 216
	    graph.node[n]['g'] = 242
	    graph.node[n]['b'] = 248
        else:
            graph.node[n]['r'] = 204
	    graph.node[n]['g'] = 255
	    graph.node[n]['b'] = 255

####################     Callable functions     ####################

def flat_mac_tree(model_file, property_list, mac_file, graphml_file, freq_bool, far_bool):
    mac_list = mac_file_2_list(mac_file)
    freq_dictionary = make_freq_dict(mac_list)    
    freq_dictionary = normalise_value(freq_dictionary, 100)
    pred_dict = makePredDict(model_file, property_list)

    # Build MAC tree  
    dg=nx.DiGraph() 
    ct = MacTree('', -1, mac_list, pred_dict, -1, [], freq_bool, far_bool)
    ct.to_di_graph(dg, None, freq_dictionary)
    
    # Color nodes
    mv=get_max_value(pred_dict)
    if mv>=10:
        norm_d = normalise_value(pred_dict, 100)
        red2blue_10grad(dg, norm_d)
    else :
        norm_d = normalise_value(pred_dict, 50)
        red2blue_5grad(dg, norm_d)
        
    # Export to graphml
    nx.write_graphml(dg,graphml_file)
    
def mac_dag(model_file, property_list, mac_file, graphml_file, freq_bool, far_bool):
    
    mac_list = mac_file_2_list(mac_file)
    freq_dictionary = make_freq_dict(mac_list)    
    freq_dictionary = normalise_value(freq_dictionary, 100)
    pred_dict = makePredDict(model_file, property_list)

    # Build MAC tree  
    dg=nx.DiGraph() 
    ct = MacTree('', -1, mac_list, pred_dict, -1, [], freq_bool, far_bool)
    
    deep_dictionary = dict()
    ct.make_deep_dict(deep_dictionary)

    # Minimize flat tree to DAG
    dg=nx.DiGraph()
    dag_list = build(deep_dictionary)
    
#    pred_dict = normalise_value(pred_dict,5)
    dg=nx.DiGraph()
    for dag in dag_list:     
        dag.to_di_graph(dg, None, freq_dictionary)
        
        # Color nodes
        mv=get_max_value(pred_dict)
        if mv>=10:
            norm_d = normalise_value(pred_dict, 100)
            red2blue_10grad(dg, norm_d)
        else :
        	norm_d = normalise_value(pred_dict, 50)
        red2blue_5grad(dg, norm_d)
        
    # Export to graphml
#    print pred_dict
    nx.write_graphml(dg,graphml_file)
    
    return dag_list

def mac_dag_minimization(list_of_dag, graphml_file, l_red, h_red, loop):
    change = True
    nb_nodes = -1
    nb_edges = -1
    while change :
        dg=nx.DiGraph()
        for dag in list_of_dag :
            dag.allocate_father()
        for dag in list_of_dag :
            if l_red :
                dag.linear_reduction()
            if h_red :
                dag.horizontal_reduction()
        for dag in list_of_dag :     
            dag.to_digraph_merge_representation(dg, None)
            
        new_nb_nodes = len(dg.nodes())
        new_nb_edges = len(dg.edges())
        print 'nodes : ',new_nb_nodes, ' edges : ',new_nb_edges
        if new_nb_nodes != nb_nodes or new_nb_edges != nb_edges :
            nb_nodes = new_nb_nodes
            nb_edges = new_nb_edges
        else :
            change = False    
        if not loop :
            change = False
            
    # Export to graphml
    nx.write_graphml(dg,graphml_file)

# For random test
def mac_dag_random_test(mac_file, graphml_file):
    # randomized mac elements
    list_of_mac = mac_file_2_list(mac_file)
    list_of_element = mac_list_2_element_list(list_of_mac)
    rand_element = make_random_list(list_of_element)

    # Build MAC tree  
    dg=nx.DiGraph() 
    ct = MacTreeRandom('', -1, list_of_mac, rand_element, -1, [])

    deep_dictionary = dict()
    ct.make_deep_dict(deep_dictionary)

    # Minimize flat tree to DAG
    dg=nx.DiGraph()
    list_of_dag = build(deep_dictionary)
    
    change = True
    nb_nodes = -1
    nb_edges = -1
    while change :
        dg=nx.DiGraph()
        for dag in list_of_dag :
            dag.allocate_father()
        for dag in list_of_dag :
            dag.linear_reduction()
            dag.horizontal_reduction()
        for dag in list_of_dag :     
            dag.to_digraph_merge_representation(dg, None)
            
        new_nb_nodes = len(dg.nodes())
        new_nb_edges = len(dg.edges())
        print 'nodes : ',new_nb_nodes, ' edges : ',new_nb_edges
        if new_nb_nodes != nb_nodes or new_nb_edges != nb_edges :
            nb_nodes = new_nb_nodes
            nb_edges = new_nb_edges
        else :
            change = False    

        change = False
        
    # Export to graphml
    nx.write_graphml(dg,graphml_file)
    
    
if __name__ == '__main__':
    
    ####################     IMPORTS     ####################
    # Import static_analysis of CADBIOM to make activation graph
    import sys
    #sys.path.append('/home/gandrieu/cadbiomworkspace/Cadbiom_des_new/') # cadbiom model path
    sys.path.append('../')
    from models.guard_transitions.analyser.static_analysis import *
    
    
    
    
    

    ####################     PARAMETERS     ####################
    
    # Inputs
    chart_model_input_file = '../bio_models/tgf/tgf_canonical_2707_AND_Clock.bcx'

    lprop = ['p21CIP1']

    mac_input_file = open('../bio_models/solution/p21reachable_tgfCanonical_cam_only.txt','r+')
    
    # Order elements
    """
    True if you order nodes by frequence and then by distance, 
    False if you order nodes by distance and then by frequence
    """
    freq_first = True
    
    """
    True if you take the farthest node first
    False if you take the nearest node first
    """
    farthest_first = True
    
    # Output
    graphml_output_file = '../bio_models/solution/p21reachable_tgfCanonical.graphml'
    
    
    
    
    #############################    CHOOSE A TEST     ###############################
    
    
    ###############################    RANDOM TEST     ###############################
    # Uncomment the following to perform random test
#    mac_dag_random_test(mac_input_file, graphml_output_file)
    ###############################    RANDOM TEST     ###############################
    
    


    ###############################     FLAT MAC TREE     ############################
    # Uncomment the following to use flat mac tree analysis
#    flat_mac_tree(chart_model_input_file, lprop, mac_input_file, graphml_output_file, freq_first, farthest_first)
    ###############################     FLAT MAC TREE     ############################
    
    
    
    
    ##################################     MAC DAG     ###############################
    # Uncomment the following to use mac dag analysis    
    dag_list = mac_dag(chart_model_input_file, lprop, mac_input_file, graphml_output_file, freq_first, farthest_first)
    
    # Minimization
#    linear_reduction = True
#    horizontale_reduction = True
#    loop_reduction = False
#    mac_dag_minimization(dag_list, graphml_output_file, linear_reduction, horizontale_reduction, loop_reduction)
    ##################################     MAC DAG     ###############################
